@@ -1,13 +1,19 @@
-"""
-MySimpleGUI
+#   __  __         ____   _                    _         ____  _   _  ___
+#  |  \/  | _   _ / ___| (_) _ __ ___   _ __  | |  ___  / ___|| | | ||_ _|
+#  | |\/| || | | |\___ \ | || '_ ` _ \ | '_ \ | | / _ \| |  _ | | | | | |
+#  | |  | || |_| | ___) || || | | | | || |_) || ||  __/| |_| || |_| | | |
+#  |_|  |_| \__, ||____/ |_||_| |_| |_|| .__/ |_| \___| \____| \___/ |___|
+#           |___/                      |_|
+#
+# See https://github.com/salabim/MySimpleGUI/blob/master/README.md for details
 
-See https://github.com/salabim/MySimpleGUI/blob/master/README.md for details
-"""
 import sys
 from pathlib import Path
 import os
+import collections
 
-version = __version__ = "1.1.6"
+version = __version__ = "1.1.10"
+mysimplegui_version = version
 
 
 class peekable:
@@ -30,6 +36,13 @@ class peekable:
             return next(self.iterator)
 
 
+registered_patches = collections.defaultdict(int)
+
+
+def register_patch(n):
+    registered_patches[n] += 1
+
+
 def splitlines(s):
     return s.splitlines()
 
@@ -55,12 +68,10 @@ class CodeList(list):
                 lines = spec.splitlines()
         else:
             lines = spec
-        if indent == 0:
-            self.extend(lines)
-        else:
-            indentstr = " " * indent
-            for line in lines:
-                self.append(indentstr + line)
+
+        indentstr = " " * indent
+        for line in lines:
+            self.append(indentstr + line.replace("\x1b", "\\x1b"))
 
 
 pysimplegui_name = "PySimpleGUI"
@@ -104,6 +115,7 @@ for line in lines:
             this_class = parts[1].split("(")[0].replace(":", "").strip()
 
     if line.strip().startswith("self.config_count = 0"):
+        register_patch(0)
         code.add(line)
         indent = line_to_indent(line)
         code.add(
@@ -112,7 +124,8 @@ self.internal = Window.Internal(self)""",
             indent=indent,
         )
 
-    if "def Read(self, timeout=None, timeout_key=TIMEOUT_KEY, close=False):" in line:  # adds attributes to Read
+    elif "def Read(self, timeout=None, timeout_key=TIMEOUT_KEY, close=False):" in line:  # adds attributes to Read
+        register_patch(1)
         indent = line_to_indent(line)
         code.add(
             """\
@@ -156,13 +169,15 @@ def __repr__(self):
         for k in sorted(self.__dict__.keys(), key=str.upper):
             v = self.__dict__[k]
             if v not in (None, (None,None)):
-                if k == "ParentContainer":
+                if k in ("ParentContainer", "ParentForm"):
                     if hasattr(v, "type_name"):
                         result.append("    " + k + " = " + v.type_name())
                     else:
                         result.append("    " + k + " = " + repr(v))
                 else:
-                    if k not in ("WindowIcon", "AllKeysDict"):
+                    if k not in ("AllKeysDict"):
+                        if k == "WindowIcon":
+                            v = str(v)[:80] + "..."
                         if isinstance(v, list):
                             v  = list_repr(v)
                         l = repr(v).split("\\n")
@@ -226,6 +241,7 @@ results = events, values
         code.add(line)  # the original return result line
 
     elif line.strip().startswith("def _RightClickMenuCallback(self, event):"):
+        register_patch(2)
         indent = line_to_indent(line)
         code.add(
             """\
@@ -248,7 +264,7 @@ def __repr__(self):
         for k in sorted(self.__dict__.keys(), key=str.upper):
             v = self.__dict__[k]
             if v not in (None, (None,None)):
-                if k == "ParentContainer":
+                if k in ("ParentContainer", "ParentForm"):
                     if hasattr(v, "type_name"):
                         result.append("    " + k + " = " + v.type_name())
                     else:
@@ -268,6 +284,7 @@ def __repr__(self):
 
         code.add(line)
     elif line.startswith("class Element"):
+        register_patch(3)
         code.add(
             """\
 class list_repr(list):
@@ -288,9 +305,23 @@ class list_repr(list):
 
         code.add(line)
     elif line.strip().startswith("return self.FindElement(key)"):  # this the original Window.__getitem__ contents:
+        register_patch(4)
         code.add(line.replace("return self.FindElement(key)", "return Window._lookup(self.AllKeysDict, key)"))
 
+    elif line.strip().startswith("if file_name"):
+        register_patch(5)
+        code.add(line.replace("file_name", "file_name or self is target_element"))
+
+    elif line.strip() == "if folder_name:":
+        register_patch(6)
+        code.add(line.replace("folder_name", "True"))
+
+    elif "target=target" in line:
+        register_patch(7)
+        code.add(line.replace("target=target", "target=target or key or k or button_text"))
+
     elif line.strip().startswith("if element.Type == ELEM_TYPE_INPUT_TEXT:"):
+        register_patch(8)
         code.add(line)
         if lines.peek().strip().startswith("try:"):
             while True:
@@ -308,17 +339,19 @@ elif element.Type == ELEM_TYPE_TEXT:
             code.add(line)
 
     elif "ELEM_TYPE_SEPARATOR):" in line:
+        register_patch(9)
         code.add(line.replace("ELEM_TYPE_SEPARATOR", "ELEM_TYPE_SEPARATOR, ELEM_TYPE_TEXT"))
 
     elif "element.Type != ELEM_TYPE_TEXT and \\" in line:
+        register_patch(10)
         pass  # remove this condition
 
     elif "if element.Key in key_dict.keys():" in line:
-        code.add(line)
+        register_patch(11)
+        #        code.add(line)
         indent = line_to_indent(line)
         code.add(
             """\
-    raise KeyError('duplicate key found in layout: ' + repr(element.Key))
 if element.Key == "AllKeysDictxx":
     raise KeyError('key "AllKeysDict" not allowed in layout')""",
             indent=indent,
@@ -327,7 +360,39 @@ if element.Key == "AllKeysDictxx":
         while line_to_indent(lines.peek()) > indent:  # remove original code
             next(lines)
 
+    elif "self.AllKeysDict = self._BuildKeyDictForWindow(self, self, dict)" in line:
+        register_patch(34)
+        indent = line_to_indent(line)
+        code.add(line)
+        code.add(
+            """\
+self.AllKeysDictnewdict = {}
+for k, v in dict.items():
+    if not(isinstance(k, tuple) and len(k) == 2 and k[0] == NONE_KEY):
+        self.AllKeysDictnewdict[k] = v
+for k, v in dict.items():
+    if isinstance(k, tuple) and len(k) == 2 and k[0] == NONE_KEY:
+        for i in itertools.count(1):
+            if i not in self.AllKeysDictnewdict:
+                k = i
+                v.Key = i
+                self.AllKeysDictnewdict[k] = v
+                break""",
+            indent=indent,
+        )
+
+    elif "element.Key = top_window.DictionaryKeyCounter" in line:
+        register_patch(35)
+        indent = line_to_indent(line)
+        code.add("element.Key = (NONE_KEY, top_window.DictionaryKeyCounter)", indent=indent)
+
+    elif line.startswith("COLOR_SYSTEM_DEFAULT = "):
+        register_patch(36)
+        code.add("NONE_KEY = object()")
+        code.add(line)
+
     elif line == "class Multiline(Element):":
+        register_patch(12)
         code.add(line)
         while not lines.peek().strip().startswith("self."):
             code.add(next(lines))
@@ -336,11 +401,13 @@ if element.Key == "AllKeysDictxx":
             """\
 self._closed = False
 self.write_fg = None
-self.write_bg = None""",
+self.write_bg = None
+self.write_font = None""",
             indent=indent,
         )
 
     elif "def write(self, txt):" in line and this_class == "Multiline":
+        register_patch(13)
 
         indent = line_to_indent(line)
         while line_to_indent(lines.peek()) > indent:
@@ -379,6 +446,7 @@ codes = (
     ("oncyan", "\x1b[46m", None, "cyan"),
     ("onwhite", "\x1b[47m", None, "white"),
     ("ondefault", "\x1b[49m", None, "ondefault"),
+    ("font", "\x1b[font", None, None),
 )
 code_fg = {}
 code_bg = {}
@@ -397,13 +465,21 @@ def write(self, s):
             if c == "\x1b":
                 for code in Multiline.code_bg:
                     if s[i:].startswith(code):
-                        break
+                        if s[i:].startswith("\x1b[font"):
+                            parts = s[i:].split('|',1)
+                            if len(parts) >1:
+                                fontspec = parts[0][6:]
+                                use = parts[0]+"|"
+                                break
+                        else:
+                            use = code
+                            break
                 else:
                     continue
                 _print_to_element(
-                    self, s[:i], sep="", end="", text_color=self.write_fg, background_color=self.write_bg
+                    self, s[:i], sep="", end="", text_color=self.write_fg, background_color=self.write_bg, font=self.write_font,
                 )
-                s = s[i + len(code) :]
+                s = s[i + len(use) :]
                 if Multiline.code_fg[code] is not None:
                     if Multiline.code_fg[code] == "default":
                         self.write_fg = None
@@ -414,9 +490,13 @@ def write(self, s):
                         self.write_bg = None
                     else:
                         self.write_bg = Multiline.code_bg[code]
+                if code == "\x1b[0m":
+                    self.write_font = None
+                if code == "\x1b[font":
+                    self.write_font = fontspec
                 break
         else:
-            _print_to_element(self, s, sep="", end="", text_color=self.write_fg, background_color=self.write_bg)
+            _print_to_element(self, s, sep="", end="", text_color=self.write_fg, background_color=self.write_bg, font=self.write_font)
             s = ""
 
 def flush(self):
@@ -435,11 +515,89 @@ def writable(self):
         )
 
     elif "element.TKText.insert(1.0, element.DefaultText)" in line:
+        register_patch(14)
         code.add(
             line.replace("element.TKText.insert(1.0, element.DefaultText)", "print(element.DefaultText, file=element)")
         )
 
+    elif "background_color_for_value=background_color" in line:
+        register_patch(15)
+        code.add(
+            line.replace(
+                "background_color_for_value=background_color",
+                "background_color_for_value=background_color, font_for_value=font",
+            )
+        )
+
+    elif "background_color_for_value=None" in line:
+        register_patch(16)
+        code.add(
+            line.replace("background_color_for_value=None", "background_color_for_value=None, font_for_value=None")
+        )
+
+    elif "if background_color_for_value is not None or text_color_for_value is not None:" in line:
+        register_patch(17)
+        code.add(line.replace(":", " or font_for_value is not None:"))
+
+    elif "str(background_color_for_value)+')'" in line:
+        register_patch(18)
+        code.add(
+            line.replace(
+                "str(background_color_for_value)+')'", "str(background_color_for_value)+',' + str(font_for_value)+')'"
+            )
+        )
+
+    elif "text_color=None, background_color=None" in line and line.strip().startswith(
+        ("def Print", "def print", "def _print_to_element")
+    ):
+        register_patch(19)
+        code.add(
+            line.replace("text_color=None, background_color=None", "text_color=None, background_color=None, font=None")
+        )
+
+    elif "text_color=text_color, background_color=background_color" in line and line.strip().startswith(
+        "_print_to_element"
+    ):
+        register_patch(20)
+        code.add(
+            line.replace(
+                "text_color=text_color, background_color=background_color",
+                "text_color=text_color, background_color=background_color, font=font",
+            )
+        )
+
+    elif "if background_color_for_value is not None:" in line:
+        register_patch(21)
+        indent = line_to_indent(line)
+        code.add(
+            """\
+if font_for_value is not None:
+    self.TKText.tag_configure(tag, font=font_for_value)""",
+            indent=indent,
+        )
+        code.add(line)
+
+    elif "if element.Key is not None:" in line:
+        register_patch(22)
+        code.add(
+            line.replace(
+                "if element.Key is not None:",
+                "if element.Key is not None and isinstance(element.Key, collections.abc.Hashable):",
+            )
+        )
+
+    elif line.strip().startswith("form.ReturnValuesDictionary[element.Key] = value"):
+        register_patch(23)
+        indent = line_to_indent(line)
+        code.add(
+            """\
+if isinstance(element.Key, collections.abc.Hashable):
+    form.ReturnValuesDictionary[element.Key] = value""",
+            indent=indent,
+        )
+
     elif line.startswith("def SetOptions("):  # generates extra function to get/set globals
+        register_patch(24)
         names = {}
         code.add(line)
         while line_to_indent(lines.peek()) > 0:
@@ -479,6 +637,7 @@ def {alias}(value = None):
             )
 
     elif line.startswith("def PopupError("):  # changes PopupError behaviour to raise exception
+        register_patch(25)
         indent = line_to_indent(line)
 
         code.add(line)
@@ -494,16 +653,18 @@ if (trace_details[-1].split(",")[0] == trace_details[-2].split(",")[0]) and RAIS
             indent=indent + 4,
         )
 
-    elif line == "import sys":  # add some required modules
+    elif line == "import queue":  # add some required modules
+        register_patch(26)
+        code.add(line)
         code.add(
             """\
 import keyword
 import collections
-import io
-import sys"""
+import io"""
         )
 
     elif line.startswith("version = "):  # check compatibilty of PySimpleGUI version (at patch time)
+        register_patch(27)
         code.add(line)
         exec(line)
         this_pysimplegui_version = version.split()[0]
@@ -525,6 +686,7 @@ import sys"""
         del minimal_pysimplegui_version
 
     elif line.strip() in ("except:", "except Exception as e:"):  # exception handler insertion
+        register_patch(28)
         indent = line_to_indent(line)
         code.add("except Exception as e:", indent=indent)
         buffered_lines = []
@@ -552,7 +714,6 @@ if RAISE_ERRORS:
 
             code.add(
                 """\
-
 if RAISE_ERRORS:
     captured = sys.stdout.getvalue()
     sys.stdout = save_stdout
@@ -563,18 +724,28 @@ if RAISE_ERRORS:
         else:
             code.add(buffered_lines)
 
+    #    elif line.strip().startswith("if key in settings.dict:"):
+    #        indent = line_to_indent(line)
+    #        while not lines.peek().strip().startswith("else:"):
+    #            code.add(next(lines).strip(), indent=indent)
+    #        while line_to_indent(next(lines)) == indent:
+    #            next(lines)
+
     elif "SUPPRESS_ERROR_POPUPS = False" in line:  # set global RAISE_ERRORS
+        register_patch(29)
         code.add(line)
         indent = line_to_indent(line)
         code.add("RAISE_ERRORS = True", indent=indent)
 
     elif "ix = random.randint(0, len(lf_values) - 1)" in line:  # no more random theme selection, but exception
+        register_patch(30)
         indent = line_to_indent(line)
         code.add("raise ValueError(index + ' not a valid theme')", indent=indent)
         while line_to_indent(lines.peek()) >= indent:
             next(lines)
 
     elif line.strip().startswith("warnings.warn("):
+        register_patch(31)
         if "popup" in lines.peek().lower() and "error" in lines.peek().lower():
             pass  # remove line
         else:
@@ -582,14 +753,88 @@ if RAISE_ERRORS:
             line = line.replace(", UserWarning", "")
             code.add(line)
 
+    elif "photo = tk.PhotoImage(file=element.Filename)" in line:
+        register_patch(32)
+        indent = line_to_indent(line)
+        code.add(
+            """\
+try:
+    import PIL
+    from PIL import ImageTk
+    from PIL import Image as PILImage
+except ModuleNotFoundError:
+    PIL = None
+if isinstance(element.Filename, (str, Path)):
+    if Path(element.Filename).is_file():
+        if PIL:
+            img = PILImage.open(element.Filename)
+            photo = ImageTk.PhotoImage(img)
+        else:
+            if Path(element.Filename).suffix.lower() in (".gif", ".png"):
+                photo = tk.PhotoImage(file=element.Filename)
+            else:
+                raise ValueError("file format not supported. Try installing PIL")
+    else:
+        raise FileNotFoundError(element.Filename)
+else:
+    if PIL:
+        photo = ImageTk.PhotoImage(element.Filename)
+    else:
+        raise ValueError("PIL image format not supported. Try installing PIL")
+""",
+            indent=indent,
+        )
+
+    elif "image = tk.PhotoImage(file=filename)" in line:
+        register_patch(33)
+        indent = line_to_indent(line)
+        code.add(
+            """\
+try:
+    import PIL
+    from PIL import ImageTk
+    from PIL import Image as PILImage
+except ModuleNotFoundError:
+    PIL = None
+if isinstance(filename, (str, Path)):
+    if Path(filename).is_file():
+        if PIL:
+            img = PILImage.open(filename)
+            image = ImageTk.PhotoImage(img)
+        else:
+            if Path(filename).suffix.lower() in (".gif", ".png"):
+                image = tk.PhotoImage(file=filename)
+            else:
+                raise ValueError("file format not supported. Try installing PIL")
+    else:
+        raise FileNotFoundError(filename)
+else:
+    if PIL:
+        image = ImageTk.PhotoImage(filename)
+    else:
+        raise ValueError("PIL image format not supported. Try installing PIL")
+
+""",
+            indent=indent,
+        )
     elif line.startswith("if __name__ == "):  # no more PySimpleGUI startup screen
+        register_patch(34)
         break
 
     else:
+        register_patch(34)
         code.add(line)
 
+missing_patches = set(range(len(registered_patches))) - set(registered_patches)
+
 for var in list(vars().keys()):
-    if not var.startswith("__") and var not in ("code", "write_filename", "os"):
+    if not var.startswith("__") and var not in (
+        "code",
+        "write_filename",
+        "os",
+        "mysimplegui_version",
+        "missing_patches",
+    ):
         del vars()[var]
 
 
@@ -606,14 +851,18 @@ if "MySimpleGUI_full_traceback" in os.environ and os.environ["MySimpleGUI_full_t
 else:
     exec("\n".join(code))
 
+if missing_patches:
+    raise Warning("not all patches applied.\n" + "Missing: " + ", ".join(map(str, missing_patches)))
+
 if __name__ == "__main__":
+
     with message_box_line_width(len(write_filename)):
         if PopupYesNo("Would you like to save the patched version to\n" + write_filename) == "Yes":
             write_file()
             Popup("Patched version saved to\n" + write_filename, title="saved")
 
-
 del var
 del code
 del write_file
 del write_filename
+del missing_patches
